/*-
 * %%----------------------------------------------------------------------------------------------
 * Archive ouverte UNIGE - AoU Access - SearchController.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2022 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.aou.controller.access;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.stream.Collectors;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.boot.autoconfigure.condition.ConditionalOnBean;
import org.springframework.core.io.FileSystemResource;
import org.springframework.data.domain.Pageable;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.ModelAttribute;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.servlet.mvc.method.annotation.StreamingResponseBody;

import ch.unige.solidify.controller.IndexResourceReadOnlyController;
import ch.unige.solidify.exception.SolidifyHttpErrorException;
import ch.unige.solidify.exception.SolidifyRuntimeException;
import ch.unige.solidify.rest.ActionName;
import ch.unige.solidify.rest.FacetPage;
import ch.unige.solidify.rest.FacetRequest;
import ch.unige.solidify.rest.FieldsRequest;
import ch.unige.solidify.rest.RestCollection;
import ch.unige.solidify.rest.SearchCondition;
import ch.unige.solidify.security.EveryonePermissions;
import ch.unige.solidify.service.IndexResourceService;
import ch.unige.solidify.util.FileTool;
import ch.unige.solidify.util.StringTool;

import ch.unige.aou.AouConstants;
import ch.unige.aou.config.AouProperties;
import ch.unige.aou.controller.AccessController;
import ch.unige.aou.model.access.PublicationDownload;
import ch.unige.aou.model.access.StatisticsInfoDto;
import ch.unige.aou.model.index.PublicationIndexEntry;
import ch.unige.aou.model.publication.DocumentFileType;
import ch.unige.aou.model.publication.MetadataFile;
import ch.unige.aou.model.rest.AouSearchCondition;
import ch.unige.aou.rest.AouActionName;
import ch.unige.aou.rest.UrlPath;
import ch.unige.aou.service.PublicationDownloadProcessingService;
import ch.unige.aou.service.PublicationDownloadService;
import ch.unige.aou.service.PublicationStatisticService;
import ch.unige.aou.service.QueryBuilderService;
import ch.unige.aou.service.SearchService;
import ch.unige.aou.service.document.CoverPageGenerator;

@RestController
@EveryonePermissions
@ConditionalOnBean(AccessController.class)
@RequestMapping(UrlPath.ACCESS_METADATA)
public class SearchController extends IndexResourceReadOnlyController<String, PublicationIndexEntry> {
  private static final Logger log = LoggerFactory.getLogger(SearchController.class);

  private static final String WITH_RESTRICTED_ACCESS_MASTERS_PARAM = "with-restricted-access-masters";

  private final String indexName;
  private final QueryBuilderService queryBuilderService;
  private final PublicationDownloadService publicationDownloadService;
  private final PublicationStatisticService publicationStatisticService;
  private final PublicationDownloadProcessingService publicationDownloadProcessingService;
  private final SearchService searchService;
  private final CoverPageGenerator coverPageGenerator;

  protected SearchController(AouProperties aouProperties, IndexResourceService<String, PublicationIndexEntry> indexResourceService,
          QueryBuilderService queryBuilderService, PublicationDownloadService archiveDownloadService,
          PublicationDownloadProcessingService publicationDownloadProcessingService,
          PublicationStatisticService publicationStatisticService,
          SearchService searchService, CoverPageGenerator coverPageGenerator) {
    super(indexResourceService);
    this.queryBuilderService = queryBuilderService;
    this.indexName = aouProperties.getIndexing().getPublishedPublicationsIndexName();
    this.publicationStatisticService = publicationStatisticService;
    this.publicationDownloadService = archiveDownloadService;
    this.publicationDownloadProcessingService = publicationDownloadProcessingService;
    this.searchService = searchService;
    this.coverPageGenerator = coverPageGenerator;
  }

  @Override
  public HttpEntity<RestCollection<PublicationIndexEntry>> list(@ModelAttribute PublicationIndexEntry search, Pageable pageable) {
    return super.list(search, pageable);
  }

  @Override
  public HttpEntity<PublicationIndexEntry> get(@PathVariable String id) {
    return super.get(id);
  }

  @GetMapping(params = "resId")
  public HttpEntity<PublicationIndexEntry> getByResId(@RequestParam(name = "resId") String resId) {
    return super.get(resId);
  }

  @GetMapping(params = "archiveId")
  public HttpEntity<PublicationIndexEntry> getByArchiveId(@RequestParam(name = "archiveId") String archiveId) {

    // Get fields to include/exclude from ES response
    FieldsRequest fieldsRequest = this.queryBuilderService.getDefaultFieldsRequest();

    AouSearchCondition archiveIdCondition = this.queryBuilderService.getFindByArchiveIdCondition(archiveId);
    FacetPage<PublicationIndexEntry> results = this.indexResourceService.search(this.getIndex(), List.of(archiveIdCondition), null,
            Pageable.ofSize(10), fieldsRequest);

    if (results.getContent().size() == 1) {
      PublicationIndexEntry t = results.stream().findFirst().get();
      this.publicationStatisticService.createViewStats(t);
      this.addLinks(t);
      return new ResponseEntity<>(t, HttpStatus.OK);
    } else if (!results.hasContent()) {
      log.info("Publication with archiveId: {} was not found", archiveId);
      return new ResponseEntity<>(HttpStatus.NOT_FOUND);
    } else {
      throw new SolidifyRuntimeException("Too many publications found");
    }
  }

  @PostMapping(value = "/" + ActionName.SEARCH)
  public HttpEntity<RestCollection<PublicationIndexEntry>> searchPost(@RequestBody List<AouSearchCondition> searchConditions,
          @RequestParam(value = WITH_RESTRICTED_ACCESS_MASTERS_PARAM, defaultValue = "false") boolean withRestrictedAccessMaster,
          Pageable pageable) {

    // Replaces field aliases by field names, set AND operator on specific fields
    List<SearchCondition> searchConditionList = this.queryBuilderService.completeSearchConditions(searchConditions);

    if (!withRestrictedAccessMaster) {
      // discard masters that are not public
      searchConditionList = this.queryBuilderService.addDiscardNonPublicMastersCondition(searchConditionList);
    }

    // Get fields to include/exclude from ES response
    FieldsRequest fieldsRequest = this.queryBuilderService.getDefaultFieldsRequest();

    // Get facets
    List<FacetRequest> facetRequests = this.queryBuilderService.getFacetsList();

    FacetPage<PublicationIndexEntry> results = this.indexResourceService.search(this.getIndex(), searchConditionList, facetRequests, pageable,
            fieldsRequest);
    return this.page2Collection(results, pageable);
  }

  @GetMapping(value = "/" + AouActionName.RUN_STORED_SEARCH + "/{storedSearchId}")
  public HttpEntity<RestCollection<PublicationIndexEntry>> runStoredSearch(@PathVariable String storedSearchId,
          @RequestParam(value = WITH_RESTRICTED_ACCESS_MASTERS_PARAM, defaultValue = "false") boolean withRestrictedAccessMaster,
          Pageable pageable) {

    List<AouSearchCondition> searchConditions = this.queryBuilderService.getStoredSearchConditions(storedSearchId);

    return this.searchPost(searchConditions, withRestrictedAccessMaster, pageable);
  }

  @GetMapping(value = "/" + AouActionName.COMBINE_STORED_SEARCHES)
  public HttpEntity<RestCollection<PublicationIndexEntry>> combineStoredResearches(
          @RequestParam(name = "searches", required = false) String commaSeparatedSearchIds,
          @RequestParam(name = "publicationIds", required = false) String commaSeparatedPublicationIds,
          @RequestParam(name = "combinationType", required = false, defaultValue = "or") String combinationType,
          @RequestParam(value = WITH_RESTRICTED_ACCESS_MASTERS_PARAM, defaultValue = "false") boolean withRestrictedAccessMaster,
          Pageable pageable) {

    if (StringTool.isNullOrEmpty(commaSeparatedSearchIds) && StringTool.isNullOrEmpty(commaSeparatedPublicationIds)) {
      throw new SolidifyHttpErrorException(HttpStatus.BAD_REQUEST, "At least one search id or one publication id is required");
    }

    List<String> searchIds = new ArrayList<>();
    if (!StringTool.isNullOrEmpty(commaSeparatedSearchIds)) {
      searchIds.addAll(Arrays.stream(commaSeparatedSearchIds.split(",")).map(String::trim).collect(Collectors.toList()));
    }

    List<String> publicationIds = new ArrayList<>();
    if (!StringTool.isNullOrEmpty(commaSeparatedPublicationIds)) {
      publicationIds.addAll(Arrays.stream(commaSeparatedPublicationIds.split(",")).map(String::trim).collect(Collectors.toList()));
    }

    AouSearchCondition searchCondition = this.queryBuilderService.getSearchesOrPublicationsSearchCondition(searchIds, publicationIds,
            combinationType);

    return this.searchPost(List.of(searchCondition), withRestrictedAccessMaster, pageable);
  }

  @GetMapping("/{id}/" + AouActionName.DOWNLOAD_STATUS)
  public HttpEntity<String> getDownloadStatus(@PathVariable String id) {
    final PublicationDownload publicationDownload = this.publicationDownloadService.findOne(id);
    if (publicationDownload == null) {
      throw new NoSuchElementException("Archive Download not found");
    }
    return new ResponseEntity<>(publicationDownload.getStatus().toString(), HttpStatus.OK);
  }

  @PostMapping("/{id}/" + AouActionName.PREPARE_DOWNLOAD)
  @PreAuthorize("@publicationDownloadPermissionService.isAllowed(#id, 'PREPARE_DOWNLOAD_ARCHIVE')")
  public HttpEntity<String> prepareDownload(@PathVariable String id) {
    try {
      this.publicationDownloadService.findOne(id);
      /*
       * Archive already download
       */
      return new ResponseEntity<>("Archive Download already exists", HttpStatus.BAD_REQUEST);
    } catch (NoSuchElementException e) {
      this.publicationDownloadService.createAndSubmitPublicationDownload(id);
      return new ResponseEntity<>(HttpStatus.ACCEPTED);
    }
  }

  @GetMapping("/{documentFileId}/" + ActionName.DOWNLOAD)
  @PreAuthorize("@publicationDownloadPermissionService.isAllowed(#documentFileId, 'DOWNLOAD_FILE') "
          + "|| @downloadTokenPermissionService.isAllowed(#documentFileId)")
  public HttpEntity<StreamingResponseBody> download(@PathVariable String documentFileId) throws IOException, URISyntaxException {
    PublicationDownload publicationDownload;
    try {
      publicationDownload = this.publicationDownloadService.findOne(documentFileId);
    } catch (NoSuchElementException e) {
      //Meaning it is coming outside the frontend
      publicationDownload = this.publicationDownloadService.createPublicationDownload(documentFileId);
      this.publicationDownloadProcessingService.processArchiveDownloadUntilReady(publicationDownload);
    }

    // get publication metadata from index
    PublicationIndexEntry indexEntry = this.indexResourceService.findOne(this.getIndex(), publicationDownload.getPublicationId());

    // Increment number of downloads (only if file has a PRINCIPAL type level and is not an Imprimatur)
    List<MetadataFile> metadataFiles = indexEntry.getFiles();
    Optional<MetadataFile> downloadedFileOpt = metadataFiles.stream().filter(f -> f.getResId().equals(documentFileId)).findFirst();
    if (downloadedFileOpt.isPresent()) {
      MetadataFile downloadedFile = downloadedFileOpt.get();
      if (!StringTool.isNullOrEmpty(downloadedFile.getTypeLevel())
              && !StringTool.isNullOrEmpty(downloadedFile.getType())
              && downloadedFile.getTypeLevel().equals(DocumentFileType.FileTypeLevel.PRINCIPAL.name())
              && !downloadedFile.getType().equals(AouConstants.DOCUMENT_FILE_TYPE_IMPRIMATUR_VALUE)) {
        this.publicationStatisticService.createDownloadStats(documentFileId, publicationDownload);
      }
    }

    final URI fileUri = publicationDownload.getDownloadUri();
    URL urlAfterCoverPageProcess = this.coverPageGenerator.generateCoverPage(documentFileId, Path.of(fileUri), indexEntry);
    InputStream is = urlAfterCoverPageProcess.openStream();

    String contentTypeFile =
            publicationDownload.getContentType() == null ? FileTool.getContentType(fileUri).toString() : publicationDownload.getContentType();

    String fileName = fileUri.getPath().substring(fileUri.getPath().lastIndexOf(File.separator) + 1);

    // If the filename isn't ISO_8859_1 compatible, rename it "file.xxx"
    CharsetEncoder iso8859Encoder = StandardCharsets.ISO_8859_1.newEncoder();
    if (!iso8859Encoder.canEncode(fileName)) {
      String dotExtension = "";
      if (fileName.contains(".")) {
        dotExtension = fileName.substring(fileName.lastIndexOf("."));
      }
      fileName = AouConstants.FILE_NAME_REPLACEMENT + dotExtension;
    }

    return this.buildDownloadResponseEntity(is, fileName, contentTypeFile, FileTool.getSize(urlAfterCoverPageProcess.toURI()));
  }

  @GetMapping("/{publicationId}/" + AouActionName.STATISTICS)
  public HttpEntity<StatisticsInfoDto> statistics(@PathVariable String publicationId) {
    return new ResponseEntity<>(this.publicationStatisticService.getPublicationStatistics(publicationId), HttpStatus.OK);
  }

  @GetMapping("/{id}/" + AouActionName.THUMBNAIL)
  public HttpEntity<FileSystemResource> downloadThumbnail(@PathVariable String id) throws IOException {
    FileSystemResource thumbnailFSR = this.searchService.downloadThumbnail(id);
    String contentTypeFile = FileTool.getContentType(thumbnailFSR.getURI()).toString();
    return this.buildFileSystemResource(thumbnailFSR, contentTypeFile, 0);
  }

  protected ResponseEntity<FileSystemResource> buildFileSystemResource(FileSystemResource inputStream, String fileContentType,
          long fileSize) {
    HttpHeaders respHeaders = new HttpHeaders();
    respHeaders.setContentType(MediaType.valueOf(fileContentType));
    if (fileSize > 0) {
      respHeaders.setContentLength(fileSize);
    }
    return new ResponseEntity<>(inputStream, respHeaders, HttpStatus.OK);
  }

  @Override
  protected String getIndex() {
    return this.indexName;
  }
}
